# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Michał Sawicz <michal@sawicz.net>
"""
Resource provider for the JustinTV plugin.
"""

import re
import simplejson

from datetime import datetime
from urllib import quote
from twisted.web2 import responsecode
from twisted.web2.stream import BufferedStream, MemoryStream

from elisa.core.common import application
from elisa.core.utils import defer
from elisa.core.media_uri import MediaUri
from elisa.core.components.resource_provider import ResourceProvider

from elisa.extern.odict import OrderedDict

from elisa.plugins.http_client.http_client import ElisaAdvancedHttpClient, \
                                                  ClientRequest, ElisaHttpClient

from elisa.plugins.base.models.media import RawDataModel
from elisa.plugins.base.models.image import ImageModel

from elisa.plugins.justintv.extern.BeautifulSoup import BeautifulSoup
from elisa.plugins.justintv.common import API_KEY, API_SECRET, API_HOST, WWW_HOST
from elisa.plugins.justintv.extern.oauth import OAuthConsumer, OAuthRequest, \
    OAuthToken, OAuthSignatureMethod_HMAC_SHA1
from elisa.plugins.justintv.channel import JustinTVChannelModel,\
    JustinTVChannelListModel
from elisa.plugins.justintv.category import JustinTVCategoryListModel
from elisa.plugins.justintv.stream import JustinTVStreamListModel
from elisa.plugins.justintv.clip import JustinTVClipListModel
from elisa.plugins.justintv.user import JustinTVUserModel


BASE_RE = "^http://(?:www\.)?%s" % re.escape(WWW_HOST)
API_RE = "^http://%s" % re.escape(API_HOST)


class JustinTVNotAuthorizedError(Exception):
    pass


class JustinTVResourceProvider(ResourceProvider):

    _access_token = None
    _request_token = None
    
    default_config = {'token_key': '', 'token_secret': ''}
    
    _uri_map = OrderedDict()
    
    # category list
    _uri_map['%s/directory/?$' % BASE_RE] = {
        'action': 'category_list',
        'uri':    'http://%s/api/category/list.json' % API_HOST,
        'client': 'api',
        'model':  JustinTVCategoryListModel,
        'auth':   False
    }
    # category stream list
    _uri_map['%s/directory/([^/]+)/?$' % BASE_RE] = {
        'action': 'stream_list',
        'uri':    'http://%s/api/stream/list.json?' \
                  'category=\\1&limit=50' % API_HOST,
        'client': 'api',
        'model':  JustinTVStreamListModel,
        'auth':   False
    }
    # subcategory stream list
    _uri_map['%s/directory/([^/]+)/([^/]+)/?$' % BASE_RE] = {
        'action': 'stream_list',
        'uri':    'http://%s/api/stream/list.json?' \
                  'subcategory=\\2&limit=50' % API_HOST,
        'client': 'api',
        'model':  JustinTVStreamListModel,
        'auth':   False
    }
    # search streams
    _uri_map['%s/search/?.*q=([^&]+)' % BASE_RE] = {
        'action': 'stream_list',
        'uri':    'http://%s/api/stream/search/\\1.json?limit=50' % API_HOST,
        'client': 'api',
        'model':  JustinTVStreamListModel,
        'auth':   False
    }
    # clips
    _uri_map['%s/([^/]+)/all?$' % BASE_RE] = {
        'action': 'clips',
        'uri':    'http://%s/api/channel/clips/\\1.json?limit=50' % API_HOST,
        'client': 'api',
        'model':  JustinTVClipListModel,
        'auth':   False
    }
    # user details
    _uri_map['%s/home/?$' % BASE_RE] = {
        'action': 'whoami',
        'uri':    'http://%s/api/account/whoami.json' % API_HOST,
        'client': 'api',
        'model':  JustinTVUserModel,
        'auth':   True
    }
    # user favorites
    _uri_map['%s/([^/]+)/favorites/?$' % BASE_RE] = {
        'action': 'favorites',
        'uri':    'http://%s/api/user/favorites/\\1.json?limit=50' % API_HOST,
        'client': 'api',
        'model':  JustinTVChannelListModel,
        'auth':   False
    }
    # add to favorites
    _uri_map['%s/([^/]+)/favorite/?$' % BASE_RE] = {
        'action': 'add_favorite',
        'uri':    'http://%s/api/fan/create/\\1.json' % API_HOST,
        'client': 'api',
        'model':  JustinTVUserModel,
        'auth':   True
    }    
    # add to favorites
    _uri_map['%s/([^/]+)/not_favorite/?$' % BASE_RE] = {
        'action': 'del_favorite',
        'uri':    'http://%s/api/fan/destroy/\\1.json' % API_HOST,
        'client': 'api',
        'auth':   True
    }    
    # channel details
    _uri_map['%s/([^/]+)/?$' % BASE_RE] = {
        'action': 'channel',
        'uri':    'http://%s/api/channel/show/\\1.json' % API_HOST,
        'client': 'api',
        'model':  JustinTVChannelModel,
        'auth':   False
    }
    # oauth request token
    _uri_map['%s/oauth/request_token$' % API_RE] = {
        'action': 'request_token',
        'client': 'api',
        'auth':   False
    }
    # oauth authenticity token
    _uri_map['%s/oauth/authorize' % BASE_RE] = {
        'action': 'auth_token',
        'client': 'www',
        'auth':   False
    }
    # oauth access token
    _uri_map['%s/oauth/access_token$' % API_RE] = {
        'action': 'access_token',
        'client': 'api',
        'auth':   False
    }
        
    supported_uri = '|'.join(_uri_map)

    _clients = {}

    def initialize(self):
        def logged_in(data):
            self._user_data = data
            return defer.succeed(self)
            
        def login_failed(error):
            self._access_token = None
            self.config['token_key'] = None
            self.config['token_secret'] = None
            return defer.succeed(self)
        
        super(JustinTVResourceProvider, self).initialize()
        self._clients['api'] = ElisaAdvancedHttpClient(host = API_HOST)
        self._clients['www'] = ElisaHttpClient(host = WWW_HOST)
        self._consumer = OAuthConsumer(API_KEY, API_SECRET)
        if self.config['token_key'] and self.config['token_secret']:
            self._access_token = OAuthToken(self.config['token_key'],
                                            self.config['token_secret'])
            dfr = self.get(MediaUri('http://%s/home' % WWW_HOST))
            dfr.addCallback(logged_in)
            dfr.addErrback(login_failed)
            return 
        else:
            return defer.succeed(self)

    def clean(self):
        dfr = defer.DeferredList([self._clients['api'].close(),
                                  self._clients['www'].close()],
                                  consumeErrors=True)

        def parent_clean(result):
            return super(JustinTVResourceProvider, self).clean()

        dfr.addCallback(parent_clean)
        return dfr

    def login(self, username=None, password=None):
        def got_access_token(token):
            self._access_token = token
            self.config['token_key'] = self._access_token.key
            self.config['token_secret'] = self._access_token.secret
            return defer.succeed(self._access_token)
        
        def got_authorization(none):
            self._access_token = self._request_token
            model, dfr = \
                self.get(MediaUri('http://%s/oauth/access_token' % API_HOST))
            dfr.addCallback(got_access_token)
            return dfr
            
        def got_authenticity_token(token):
            params = {'authorize': 1,
                      'authenticity_token': token,
                      'oauth_token': self._request_token.key,
                      'user[login]': username,
                      'user[password]': password}
            model, dfr = self.post(MediaUri('http://%s/oauth/authorize' % 
                                            WWW_HOST),
                                   **params)
            dfr.addCallback(got_authorization)
            return dfr
            
        def got_request_token(token):
            self._request_token = token
            model, dfr = \
                self.get(MediaUri('http://%s/oauth/authorize?oauth_token=%s' % 
                                  (WWW_HOST, token.key)))
            dfr.addCallback(got_authenticity_token)
            return dfr
            
        model, dfr = \
            self.get(MediaUri('http://%s/oauth/request_token' % API_HOST))
        dfr.addCallback(got_request_token)
        return dfr

    def get_signed_request(self, request):
        request.sign_request(OAuthSignatureMethod_HMAC_SHA1(), 
                             self._consumer, self._access_token)
        if request.http_method == 'GET':
            headers = {}
            for k, v in request.to_header().items():
                headers[k] = [v]
            return ClientRequest('GET', request.get_normalized_http_url(),
                                 headers, None)
        else:
            return ClientRequest('POST', request.get_normalized_http_url(),
                                 None, MemoryStream(request.to_postdata()))

    def map_uri(self, uri, context_model=None):
        url = str(uri)
        # select the correct HTTP client and target model
        for (regex, item) in self._uri_map.items():
            # precompile the regex
            if not item.has_key('re'):
                item['re'] = re.compile(regex)
            match = item['re'].match(url)
            if match is not None:
                if item['auth'] and not self._access_token:
                    return defer.fail(JustinTVNotAuthorizedError())
                http_client = self._clients[item['client']]
                action = item['action']
                if item.has_key('model'):
                    if isinstance(context_model, item['model']):
                        # reuse the context model
                        result_model = context_model
                    else:
                        # instantiate the relevant model
                        result_model = item['model']()
                else:
                    result_model = None
                # rewrite the uri
                if item.has_key('uri'):
                    uri = MediaUri(item['re'].sub(item['uri'], url))
                    self.debug("Rewritten GET %s" % uri)
                break
        return http_client, action, result_model, uri

    def get(self, uri, context_model = None):
        """
        GET request to the JustinTV servers.
        """
        self.debug('GET %s' % uri)
        
        http_client, action, result_model, uri = self.map_uri(uri, context_model)
        
        def response_read(response, model):
            if action in ['category_list', 'stream_list', 'clips', 'whoami',
                          'favorites', 'channel']:
                data = simplejson.loads(response)
                if isinstance(data, dict) and data.has_key('error'):
                    return defer.fail(Exception(data['error']['field']))
                result_model.populate(data)
                return result_model
            elif action == 'whoami':
                return simplejson.loads(response)
            elif action in ['request_token', 'access_token']:
                return OAuthToken.from_string(response)
            elif action == 'auth_token':
                soup = BeautifulSoup(response)
                form = soup.find('form', action='/oauth/authorize')
                input = form.find('input', attrs={'name': 'authenticity_token'}) 
                return input.get('value')

        def request_done(response, model):
            if response.code == responsecode.OK:
                # Read the response stream
                read_dfr = BufferedStream(response.stream).readExactly()
                read_dfr.addCallback(response_read, model)
                return read_dfr
            elif response.code == responsecode.NOT_FOUND:
                # 404 error code: resource not found
                return defer.fail(IOError('Resource not found at %s' % uri))
            else:
                # Other HTTP response code
                return defer.fail(Exception('Received an %d HTTP response code' \
                                            % response.code))
        
        if http_client == self._clients['api']:
            url = '%s://%s%s' % (uri.scheme, uri.host, uri.path)
            request = \
                OAuthRequest.from_consumer_and_token(self._consumer,
                                                     self._access_token,
                                                     http_method='GET',
                                                     http_url=url,
                                                     parameters=uri.get_params())
            request = self.get_signed_request(request)
        else:
            resource = str('%s?%s' % (uri.path, uri.get_params_string()))
            request = ClientRequest('GET', resource, None, None)

        request_dfr = http_client.request_full(request)
        request_dfr.addCallback(request_done, result_model)
        return (None, request_dfr)
    
    def post(self, uri, **parameters):
        """
        POST request to the JustinTV servers.
        """

        self.debug('POST %s' % uri)
        
        http_client, action, result_model, uri = self.map_uri(uri)

        def request_done(response):
            if response.code == responsecode.OK \
            or response.code == responsecode.FOUND:
                return defer.succeed(None)
            elif response.code == responsecode.NOT_FOUND:
                # 404 error code: resource not found
                return defer.fail(IOError('Resource not found at %s' % uri))
            else:
                # Other HTTP response code
                return defer.fail(Exception('Received an %d HTTP response code'
                                            % response.code))
        
        if http_client == self._clients['api']:
            url = '%s://%s%s' % (uri.scheme, uri.host, uri.path)
            request = OAuthRequest.from_consumer_and_token(self._consumer,
                                                           self._access_token,
                                                           http_method='POST',
                                                           http_url=url,
                                                           parameters=parameters)
            request = self.get_signed_request(request)
        else:
            params = \
                '&'.join(['%s=%s' % (quote(str(k), '~'), quote(str(v), '~')) \
                for k, v in parameters.iteritems()])
            resource = str('%s?%s' % (uri.path, uri.get_params_string()))
            request = ClientRequest('POST', resource, None, MemoryStream(params))

        request_dfr = http_client.request_full(request)
        request_dfr.addCallback(request_done)
        return (None, request_dfr)
    
    def logged_in(self):
        if self._access_token:
            return True
        return False