# -*- coding: utf-8 -*-
# Moovida - Home multimedia server
# Copyright (C) 2006-2009 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Moovida with Fluendo's plugins.
#
# The GPL part of Moovida is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Moovida" in the root directory of this distribution package
# for details on that license.
#
# Author: Michał Sawicz <michal@sawicz.net>
"""
Unit tests for the JustinTV resource provider.
"""

import binascii
import time

from urllib import quote
from twisted.trial.unittest import TestCase

from elisa.core.utils.misc import run_functional_tests_check
from elisa.core.media_uri import MediaUri

from elisa.plugins.base.models.media import PlayableModel, RawDataModel

from elisa.plugins.justintv.extern.oauth import OAuthToken
from elisa.plugins.justintv.resource_provider import JustinTVResourceProvider, \
    WWW_HOST, API_HOST, JustinTVNotAuthorizedError
from elisa.plugins.justintv.category import JustinTVCategoryListModel,\
    JustinTVCategoryModel
from elisa.plugins.justintv.stream import JustinTVStreamModel,\
    JustinTVStreamListModel
from elisa.plugins.justintv.clip import JustinTVClipModel, JustinTVClipListModel
from elisa.plugins.justintv.channel import JustinTVChannelModel,\
    JustinTVChannelListModel
from elisa.plugins.justintv.archive import JustinTVArchiveListModel,\
    JustinTVArchiveModel
from elisa.plugins.justintv.user import JustinTVUserModel

class TestJustinTVResourceProvider(TestCase):

    """
    This test case tests the JustinTVResourceProvider class.
    """

    def setUp(self):
        run_functional_tests_check()
        def created(provider):
            self.provider = provider
            return provider.initialize()

        create_dfr = JustinTVResourceProvider.create({}).addCallback(created)
        return create_dfr

    def tearDown(self):
        return self.provider.clean()

    def test_get_wrong_uri(self):
        """
        Test trying to get an unexisting URI.
        """
        uri = 'http://%s/%s' % (WWW_HOST, 'crap')
        result_model, request_defer = self.provider.get(MediaUri(uri))
        self.failUnlessFailure(request_defer, Exception)
        return request_defer
    
    def test_get_category_list(self):
        """
        Test getting the category and subcategory list
        """
        def got_categories(result_model):
            self.failUnlessIsInstance(result_model, JustinTVCategoryListModel)
            self.failUnless(result_model)
            self.failUnless(reduce(lambda x, y: x and isinstance(y, JustinTVCategoryModel), result_model))
            return result_model
        
        uri = 'http://%s/directory' % WWW_HOST
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(got_categories)
        return dfr

    def test_get_category_stream_list(self):
        """
        Test getting a list of streams assigned to a category
        """
        def got_streams(result_model):
            self.failUnlessIsInstance(result_model, JustinTVStreamListModel)
            self.failUnless(result_model)
            self.failUnless(reduce(lambda x, y: x and isinstance(y, JustinTVStreamModel), result_model))
            return result_model
        
        uri = 'http://%s/directory/%s' % (WWW_HOST, 'featured')
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(got_streams)
        return dfr

    def test_search(self):
        """
        Test searching for streams
        """
        def got_streams(result_model):
            self.failUnlessIsInstance(result_model, JustinTVStreamListModel)
            self.failUnless(result_model)
            self.failUnless(reduce(lambda x, y: x and isinstance(y, JustinTVStreamModel), result_model))
            return result_model
        
        uri = 'http://%s/search/?q=%s' % (WWW_HOST, quote('hd office'))
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(got_streams)
        return dfr

    def test_get_channel_clip_list(self):
        """
        Test getting a list of clips for a channel
        """
        def got_clips(result_model):
            self.failUnlessIsInstance(result_model, JustinTVClipListModel)
            self.failUnless(result_model)
            self.failUnless(reduce(lambda x, y: x and isinstance(y, JustinTVClipModel), result_model))
            return result_model
        
        uri = 'http://%s/%s/all' % (WWW_HOST, 'petersanderstead')
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(got_clips)
        return dfr

    def test_get_channel_archive_list(self):
        """
        Test getting a list of archives for a channel
        """
        def got_archives(result_model):
            self.failUnlessIsInstance(result_model, JustinTVArchiveListModel)
            self.failUnless(result_model)
            self.failUnless(reduce(lambda x, y: x and isinstance(y, JustinTVArchiveModel), result_model))
            return result_model
        
        uri = 'http://%s/%s/archive' % (WWW_HOST, 'petersanderstead')
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(got_archives)
        return dfr

    def test_get_favorites(self):
        """
        Test getting a list of favorite channels
        """
        def got_favorites(result_model):
            self.failUnlessIsInstance(result_model, JustinTVChannelListModel)
            self.failUnless(result_model)
            self.failUnless(reduce(lambda x, y: x and isinstance(y, JustinTVChannelModel), result_model))
            return result_model
        
        uri = 'http://%s/%s/favorites' % (WWW_HOST, 'justin')
        result_model, dfr = self.provider.get(MediaUri(uri))
        dfr.addCallback(got_favorites)
        return dfr

#    def test_login(self):
#        """
#        Test logging in
#        """
#        def logged_in(provider):
#            self.provider._access_token = None
#            return provider
#        
#        dfr = self.provider.login('username', 'password')
#        dfr.addCallback(logged_in)
#        return dfr

    def test_wrong_login(self):
        """
        Test that wrong login fails correctly
        """
        dfr = self.provider.login('bad', 'login')
        self.failUnlessFailure(dfr, JustinTVNotAuthorizedError)
        return dfr
        
#    def test_access(self):
#        """
#        Test access token access
#        """
#        def got_whoami(data):
#            self.failUnlessIsInstance(data, JustinTVUserModel)
#            self.provider._access_token = None
#            return data
#                        
#        uri = 'http://%s/home' % WWW_HOST
#        token_key = 'key'
#        token_secret = 'secret'
#        self.provider._access_token = OAuthToken(token_key, token_secret) 
#        model, dfr = self.provider.get(MediaUri(uri))
#        dfr.addCallback(got_whoami)
#        return dfr

    def test_no_access(self):
        """
        Test that trying to access a protected resource without
        authentication fails 
        """
        uri = 'http://%s/home' % WWW_HOST
        model, dfr = self.provider.get(MediaUri(uri))
        self.failUnlessFailure(dfr, JustinTVNotAuthorizedError)
        return dfr

#    def test_add_favorite(self):
#        """
#        Test adding a favorite
#        """
#        def added_favorite(data):
#            self.provider._access_token = None
#            return data
#                        
#        token_key = 'key'
#        token_secret = 'secret'
#        self.provider._access_token = OAuthToken(token_key, token_secret) 
#
#        uri = 'http://%s/%s/favorite' % (WWW_HOST, 'officecam')
#        model, dfr = self.provider.post(MediaUri(uri))
#        dfr.addCallback(added_favorite)
#        
#        return dfr
#
#    def test_del_favorite(self):
#        """
#        Test deleting a favorite
#        """
#        def removed_favorite(data):
#            self.provider._access_token = None
#            return data
#
#        token_key = 'key'
#        token_secret = 'secret'
#        self.provider._access_token = OAuthToken(token_key, token_secret) 
#        
#        # wait for the favorite to get added first
#        time.sleep(5)
#
#        uri = 'http://%s/%s/not_favorite' % (WWW_HOST, 'officecam')
#        model, dfr = self.provider.post(MediaUri(uri))
#        dfr.addCallback(removed_favorite)
#
#        return dfr
